// -----------------------------------------
// gid - Return the element with 'id'
// -----------------------------------------

function gid(id)
{
   return(document.getElementById(id));
}

// -----------------------------------------
// makeCell - Create a cell with 'id'
// -----------------------------------------

function makeCell(id)
{
   return(
      {
         mID    : id,
         mNorth : null,
         mSouth : null,
         mEast  : null,
         mWest  : null
      }
   );
}

// -----------------------------------------
// makeGrid - Create a 2d array of 'rows'
//            by 'columns' cells.
// -----------------------------------------

function makeGrid(rows, columns)
{
   var theGrid = [];
   var k = 0;

   for(var i = 0; i < rows; i ++)
   {
      var row = [];

      for(var j = 0; j < columns; j ++)
         row.push(makeCell(k++));

      theGrid.push(row);
   }

   return(theGrid);
}

// -----------------------------------------
// bgColor - Return our background color
// -----------------------------------------

function bgColor()
{
   // It'd be better if we could pull this value
   // straight from one of our buttons rather
   // than hard-coding it

   return("#008cba");
}

// -----------------------------------------
// startColor - Color of the first cell
// -----------------------------------------

function startColor()
{
   return("#aaffaa");
}

// -----------------------------------------
// endColor - Color of the last cell
// -----------------------------------------

function endColor()
{
   return("#ffaaaa");
}

// -----------------------------------------
// drawPiece - Render our piece at its
//             current location
// -----------------------------------------

function drawPiece(theObject, size, cellSize, which, color)
{
   var xc = theObject.mXOffset + which.column * size + size / 2;
   var yc = theObject.mYOffset + which.row    * size + size / 2;

   var oldStyle = theObject.mContext.fillStyle;

   theObject.mContext.fillStyle = color;

   theObject.mContext.fillRect(
      xc - (cellSize / 2), yc - (cellSize / 2), cellSize, cellSize
   );

   theObject.mContext.fillStyle = oldStyle;
}

// -----------------------------------------
// drawStartEnd - Render our first and
//                last cells
// -----------------------------------------

function drawStartEnd(size, theObject)
{
   var rs = theObject.mStartEnd.mRowStart;
   var cs = theObject.mStartEnd.mColumnStart;
   var re = theObject.mStartEnd.mRowEnd;
   var ce = theObject.mStartEnd.mColumnEnd;

   var cellSize = .8 * size;

   drawPiece(
      theObject,
      size,
      cellSize,
      { row: rs, column: cs },
      startColor()
   );

   drawPiece(
      theObject,
      size,
      cellSize,
      { row: re, column: ce },
      endColor()
   );
}

// -----------------------------------------
// doClear - Clear the drawing canvas
// -----------------------------------------

function doClear(context)
{
   var canvas  = gid("the-canvas");

   context.clearRect(0, 0, canvas.width, canvas.height);
   context.beginPath();
}

// -----------------------------------------
// showGrid - Draw our maze
// -----------------------------------------

function showGrid(theObject)
{
   var xOffset = theObject.mXOffset;
   var yOffset = theObject.mYOffset;
   var size    = theObject.mSize;

   var draw = function(x0, y0, x1, y1)
   {
      theObject.mContext.moveTo(xOffset + x0, yOffset + y0);
      theObject.mContext.lineTo(xOffset + x1, yOffset + y1);
      theObject.mContext.stroke();
   };

   var cellSize = (size / 4);

   if(theObject.mPrevious == null)
   {
      doClear(theObject.mContext);

      for(var i = 0; i < theObject.mGrid.length; i ++)
      {
         var y0 = i  * size;
         var y1 = y0 + size;

         for(var j = 0; j < theObject.mGrid[i].length; j ++)
         {
            var x0 = j  * size;
            var x1 = x0 + size;

            if(theObject.mGrid[i][j].mSouth == null)
               draw(x0, y1, x1, y1);

            if(theObject.mGrid[i][j].mNorth == null)
               draw(x0, y0, x1, y0);

            if(theObject.mGrid[i][j].mEast == null)
               draw(x1, y0, x1, y1);

            if(theObject.mGrid[i][j].mWest == null)
               draw(x0, y0, x0, y1);
         }
      }

      drawStartEnd(size, theObject);
   }
   else
   {
      var pWasStart =
         (theObject.mPrevious.row    == theObject.mStartEnd.mRowStart &&
          theObject.mPrevious.column == theObject.mStartEnd.mColumnStart)

      var pWasEnd =
         (theObject.mPrevious.row    == theObject.mStartEnd.mRowEnd &&
          theObject.mPrevious.column == theObject.mStartEnd.mColumnEnd)

      var color = "#ffffff";

      if(pWasStart)
         color = startColor();

      else if(pWasEnd)
         color = endColor();

      drawPiece(
         theObject, size, cellSize + 1, theObject.mPrevious, color
      );
   }

   drawPiece(theObject, size, cellSize, theObject.mCurrent, bgColor());

   // If we just landed on "end", say so

   if(theObject.mCurrent.row    == theObject.mStartEnd.mRowEnd &&
      theObject.mCurrent.column == theObject.mStartEnd.mColumnEnd)
   {
      alert("Victory!");
   }

   return(theObject);
}

// -----------------------------------------
// getRandomNumber - Return a random value
// -----------------------------------------

function getRandomNumber(maxValue)
{
   return(Math.floor(maxValue * Math.random()));
}

// -----------------------------------------
// rcKey - Construct a unique key string
//         from 'row' and 'column'
// -----------------------------------------

function rcKey(row, column)
{
   return(row + "_" + column);
}

// -----------------------------------------
// show - Show 'object' in an alert box
// -----------------------------------------

function show(object)
{
   alert(JSON.stringify(object));
}

// -----------------------------------------
// traverse - Recursively visit every cell
//            on the board that we've not
//            already visited.  Invoke
//            'callback' if there are no
//            cells remaining we can visit
// -----------------------------------------

function traverse(callback, row, column, grid, flags, count)
{
   flags[rcKey(row, column)] = true;

   var found = false;

   // Any directions we can follow?

   if(grid[row][column].mWest != null && !flags[rcKey(row, column - 1)])
   {
      found = true;
      traverse(callback, row, column - 1, grid, flags, count + 1);
   }

   if(grid[row][column].mEast != null && !flags[rcKey(row, column + 1)])
   {
      found = true;
      traverse(callback, row, column + 1, grid, flags, count + 1);
   }

   if(grid[row][column].mNorth != null && !flags[rcKey(row - 1, column)])
   {
      found = true;
      traverse(callback, row - 1, column, grid, flags, count + 1);
   }

   if(grid[row][column].mSouth != null && !flags[rcKey(row + 1, column)])
   {
      found = true;
      traverse(callback, row + 1, column, grid, flags, count + 1);
   }

   if(!found)
      callback(flags, row, column, count);

   flags[rcKey(row, column)] = false;
}

// -----------------------------------------
// getFlags - Create our 'visited' flags
//            object from 'grid'
// -----------------------------------------


function getFlags(grid)
{
   var flags  = {};

   for(var i = 0; i < grid.length; i ++)
      for(var j = 0; j < grid[i].length; j ++)
         flags[rcKey(i, j)] = false;

   return(flags);
}

// -----------------------------------------
// getStartEnd - Find our starting and
//               ending cells
// -----------------------------------------

function getStartEnd(grid)
{
   // Find the most distant cell (in terms of number of moves)
   // from our starting cell

   var row    = 0;
   var column = 0;
   var flags  = getFlags(grid);

   var maxCount  = 0;
   var maxRow    = 0;
   var maxColumn = 0;
   var maxFlags  = null;

   var callback = function(flags, row, column, count)
   {
      if(count > maxCount)
      {
         maxCount  = count;
         maxRow    = row;
         maxColumn = column;
         maxFlags  = JSON.parse(JSON.stringify(flags));
      }
   };

   traverse(callback, row, column, grid, flags, 0);

   // Now, run it again using maxRow and maxColumn as the
   // starting location

   row      = maxRow;
   column   = maxColumn;
   maxCount = 0;

   traverse(callback, row, column, grid, flags, 0);

   return(
      {
         mRowStart   : row,
         mColumnStart: column,
         mRowEnd     : maxRow,
         mColumnEnd  : maxColumn,
         mFlags      : maxFlags,
         mCount      : maxCount
      }
   );
}

// -----------------------------------------
// getRC - Get the current size setting
// -----------------------------------------

function getRC()
{
   var rc    = document.forms['form'].rows_and_columns;
   var value = rc.options[rc.selectedIndex].value;

   return(
      { rows: value, columns: value }
   );
}

// -----------------------------------------
// maze - Create and return our maze object
// -----------------------------------------

function maze(rc)
{
   var canvas  = gid("the-canvas");
   var context = canvas.getContext("2d");

   var size =
      Math.min(canvas.width, canvas.height) /
      Math.max(rc.rows, rc.columns);

   var grid = makeGrid(rc.rows, rc.columns);

   var xOffset = (canvas.width  - rc.columns * size) / 2;
   var yOffset = (canvas.height - rc.rows    * size) / 2;

   // Create our array of walls

   var walls = [];

   for(var i = 0; i < grid.length; i ++)
   {
      for(var j = 0; j < grid[i].length; j ++)
      {
         if(i > 0)
         {
            walls.push(
               { first   : grid[i - 1][j],
                 second  : grid[i][j],
                 vertical: false
               }
            );
         }

         if(j > 0)
         {
            walls.push(
               {
                  first   : grid[i][j - 1],
                  second  : grid[i][j],
                  vertical: true
               }
            );
         }
      }
   }

   //alert(JSON.stringify(walls));

   // Shuffle the walls as we want a
   // different maze every time

   for(var i = walls.length - 1; i >= 0; i --)
   {
      var j = getRandomNumber(i + 1);

      if(i != j)
      {
         var tmp  = walls[j];
         walls[j] = walls[i];
         walls[i] = tmp;
      }
   }

   //alert(JSON.stringify(walls));
   
   // For each wall, see if the two rooms the wall divides
   // are already connected

   for(var i = 0; i < walls.length; i ++)
   {
      var id1 = walls[i].first.mID;
      var id2 = walls[i].second.mID;

      if(id1 != id2)
      {
         if(walls[i].vertical)
            walls[i].first.mEast = walls[i].second.mWest = {};

         else
            walls[i].first.mSouth = walls[i].second.mNorth = {};

         for(var j = 0; j < grid.length; j ++)
            for(var k = 0; k < grid[j].length; k ++)
               if(grid[j][k].mID == id1)
                  grid[j][k].mID = id2;
      }
   }

   var startEnd = getStartEnd(grid);

   return(
      {
         mXOffset : xOffset,
         mYOffset : yOffset,
         mSize    : size,
         mGrid    : grid,
         mContext : context,
         mPrevious: null,
         mCurrent :
         {
            row   : startEnd.mRowStart,
            column: startEnd.mColumnStart
         },
         mStartEnd: startEnd
      }
   );
}

// -----------------------------------------
// doMove - Attempt to move in the specified
//          direction.
// -----------------------------------------

function doMove(rowMove, columnMove)
{
   var cr = theObject.mCurrent.row;
   var cc = theObject.mCurrent.column;

   if(rowMove > 0)
   {
      if(theObject.mGrid[cr][cc].mSouth == null)
         return;
   }
   else if(rowMove < 0)
   {
      if(theObject.mGrid[cr][cc].mNorth == null)
         return;
   }
   else if(columnMove > 0)
   {
      if(theObject.mGrid[cr][cc].mEast == null)
         return;
   }
   else
   {
      if(theObject.mGrid[cr][cc].mWest == null)
         return;
   }

   if(theObject.mPrevious == null)
      theObject.mPrevious = {};

   theObject.mPrevious.row    = theObject.mCurrent.row;
   theObject.mPrevious.column = theObject.mCurrent.column;

   theObject.mCurrent.row    += rowMove;
   theObject.mCurrent.column += columnMove;

   showGrid(theObject);
}

// -----------------------------------------
// setArrowKeyHandler - Initialize the
//                      handler so that our
//                      'doMove' function is
//                      called when an arrow
//                      key is pressed.
// -----------------------------------------

function setArrowKeysHandler()
{
   document.onkeydown = function(event)
   {
      if(!event)
         event = window.event;

      var code = event.keyCode;

      if(event.charCode && code == 0)
         code = event.charCode;

      switch(code)
      {
         case 37: doMove(0, -1); break;
         case 38: doMove(-1, 0); break;
         case 39: doMove(0, 1);  break;
         case 40: doMove(1, 0);  break;
      }

      event.preventDefault();
   };
}

// -----------------------------------------
// getNewGame - Get and render our maze
// -----------------------------------------

function getNewGame()
{
   return(
      showGrid(maze(getRC()))
   )
}

// -----------------------------------------
// newGame - Initialize 'theObject' our
//           game object
// -----------------------------------------

function newGame()
{
   theObject = getNewGame()
}
